from __future__ import annotations


import PySide6 # noqa: F401 (need to import to select the qt backend)
from PySide6 import QtWidgets, QtGui, QtCore

from wgpu.gui.qt import WgpuCanvas, call_later
from . import VisualizerCanvasBase
from ..drawreason import DrawReason
from ..recorder import VisualizationRecorder

import os
import time
import logging
import matplotlib as mpl

from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from ..visualizer import Visualizer


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

def _get_icon(name):
    this_dir = os.path.dirname(os.path.abspath(__file__))
    return QtGui.QIcon(os.path.join(this_dir, "icons", name))


class MyLineEdit(QtWidgets.QLineEdit):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._timer = QtCore.QTimer()
        self._timer.setSingleShot(True)
        self._timer.timeout.connect(self.selectAll)

    def focusInEvent(self, event):
        super().focusInEvent(event)
        self._timer.start(0)

class RecordingSettingsDialog(QtWidgets.QDialog):

    def __init__(self, *args):
        super().__init__(*args)
        self.setWindowTitle("Recording settings")
        self._layout = QtWidgets.QVBoxLayout()
        self.setLayout(self._layout)

        # checkbox for smoothing:
        self._smooth_checkbox = QtWidgets.QCheckBox("Smooth timestream camera movements")
        self._smooth_checkbox.setChecked(True)
        self._layout.addWidget(self._smooth_checkbox)

        # leave some space:
        self._layout.addSpacing(10)

        # checkbox for including vmin/vmax:
        self._vmin_vmax_checkbox = QtWidgets.QCheckBox("Set vmin/vmax from timestream")
        self._vmin_vmax_checkbox.setChecked(True)
        self._layout.addWidget(self._vmin_vmax_checkbox)

        # checkbox for changing quantity:
        self._quantity_checkbox = QtWidgets.QCheckBox("Set quantity from timestream")
        self._quantity_checkbox.setChecked(True)
        self._layout.addWidget(self._quantity_checkbox)

        self._layout.addSpacing(10)

        # checkbox for showing colorbar:
        self._colorbar_checkbox = QtWidgets.QCheckBox("Show colorbar")
        self._colorbar_checkbox.setChecked(True)
        self._layout.addWidget(self._colorbar_checkbox)

        # checkbox for showing scalebar:
        self._scalebar_checkbox = QtWidgets.QCheckBox("Show scalebar")
        self._scalebar_checkbox.setChecked(True)
        self._layout.addWidget(self._scalebar_checkbox)

        self._layout.addSpacing(10)


        # select resolution from dropdown, with options half HD, HD, 4K
        self._resolution_dropdown = QtWidgets.QComboBox()
        self._resolution_dropdown.addItems(["Half HD (960x540)", "HD (1920x1080)", "4K (3840x2160)"])
        self._resolution_dropdown.setCurrentIndex(1)

        # select fps from dropdown, with options 24, 30, 60
        self._fps_dropdown = QtWidgets.QComboBox()
        self._fps_dropdown.addItems(["24 fps", "30 fps", "60 fps"])
        self._fps_dropdown.setCurrentIndex(1)

        # put resolution/fps next to each other horizontally:
        self._resolution_fps_layout = QtWidgets.QHBoxLayout()
        self._resolution_fps_layout.addWidget(self._resolution_dropdown)
        self._resolution_fps_layout.addWidget(self._fps_dropdown)
        self._layout.addLayout(self._resolution_fps_layout)

        self._layout.addSpacing(10)

        # cancel and save.. buttons:
        self._cancel_save_layout = QtWidgets.QHBoxLayout()
        self._cancel_button = QtWidgets.QPushButton("Cancel")
        self._cancel_button.clicked.connect(self.reject)
        self._save_button = QtWidgets.QPushButton("Save")
        # save button should be default:
        self._save_button.setDefault(True)
        self._save_button.clicked.connect(self.accept)
        self._cancel_save_layout.addWidget(self._cancel_button)
        self._cancel_save_layout.addWidget(self._save_button)
        self._layout.addLayout(self._cancel_save_layout)

        # show as a sheet on macos:
        #self.setWindowModality(QtCore.Qt.WindowModality.WindowModal)
        self.setWindowFlags(QtCore.Qt.WindowType.Sheet)

    @property
    def fps(self):
        return float(self._fps_dropdown.currentText().split()[0])

    @property
    def resolution(self):
        import re
        # use regexp
        # e.g. the string 'blah (123x456)' should map to tuple (123,456)
        match = re.match(r".*\((\d+)x(\d+)\)", self._resolution_dropdown.currentText())
        return int(match.group(1)), int(match.group(2))

    @property
    def smooth(self):
        return self._smooth_checkbox.isChecked()

    @property
    def set_vmin_vmax(self):
        return self._vmin_vmax_checkbox.isChecked()

    @property
    def set_quantity(self):
        return self._quantity_checkbox.isChecked()

    @property
    def show_colorbar(self):
        return self._colorbar_checkbox.isChecked()

    @property
    def show_scalebar(self):
        return self._scalebar_checkbox.isChecked()







class VisualizationRecorderWithQtProgressbar(VisualizationRecorder):

    def __init__(self, visualizer: Visualizer, parent_widget: QtWidgets.QWidget):
        super().__init__(visualizer)
        self._parent_widget = parent_widget

    def _progress_iterator(self, ntot):
        progress_bar = QtWidgets.QProgressDialog("Rendering to mp4...", "Stop", 0, ntot, self._parent_widget)
        progress_bar.setWindowModality(QtCore.Qt.WindowModality.WindowModal)
        progress_bar.forceShow()

        last_update = 0

        loop = QtCore.QEventLoop()

        try:
            for i in range(ntot):
                # updating the progress bar triggers a render in the main window, which
                # in turn is quite slow (because it can trigger software rendering
                # of resizable elements like the colorbar). So only update every half second or so.
                if time.time() - last_update > 0.5:
                    last_update = time.time()
                    progress_bar.setValue(i)

                    with self._visualizer.prevent_sph_rendering():
                        loop.processEvents(QtCore.QEventLoop.ProcessEventsFlag.AllEvents)

                    if progress_bar.wasCanceled():
                        break
                yield i

        finally:
            progress_bar.close()

class VisualizerCanvas(VisualizerCanvasBase, WgpuCanvas):
    _default_quantity_name = "Projected density"
    _all_instances = []
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._all_instances.append(self)
        self.hide()

        self._toolbar = QtWidgets.QToolBar()
        self._toolbar.setIconSize(QtCore.QSize(16, 16))

        # setup toolbar to show text and icons
        self._toolbar.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon)

        self._load_icons()

        self._record_action = QtGui.QAction(self._record_icon, "Record", self)
        self._record_action.triggered.connect(self.on_click_record)

        self._save_action = QtGui.QAction(self._save_icon, "Snapshot", self)
        self._save_action.triggered.connect(self.on_click_save)

        self._save_movie_action = QtGui.QAction(self._save_movie_icon, "Save mp4", self)
        self._save_movie_action.triggered.connect(self.on_click_save_movie)
        self._save_movie_action.setDisabled(True)

        self._save_script_action = QtGui.QAction(self._export_icon, "Save timestream", self)
        self._save_script_action.triggered.connect(self.on_click_save_script)
        self._save_script_action.setDisabled(True)

        self._load_script_action = QtGui.QAction(self._import_icon, "Load timestream", self)
        self._load_script_action.triggered.connect(self.on_click_load_script)

        self._link_action = QtGui.QAction(self._unlinked_icon, "Link to other windows", self)
        self._link_action.setIconText("Link")
        self._link_action.triggered.connect(self.on_click_link)


        self._colormap_menu = QtWidgets.QComboBox()
        self._colormap_menu.addItems(mpl.colormaps.keys())
        self._colormap_menu.setCurrentText(self._visualizer.colormap_name)
        self._colormap_menu.currentTextChanged.connect(self._colormap_menu_changed_action)

        self._quantity_menu = QtWidgets.QComboBox()
        self._quantity_menu.addItem(self._default_quantity_name)
        self._quantity_menu.setEditable(True)



        self._quantity_menu.setLineEdit(MyLineEdit())

        # at this moment, the data loader hasn't been initialized yet, so we can't
        # use it to populate the menu. This needs a callback:
        def populate_quantity_menu():
            self._quantity_menu.addItems( self._visualizer.data_loader.get_quantity_names())
            self._quantity_menu.setCurrentText(self._visualizer.quantity_name or self._default_quantity_name)
            self._quantity_menu.currentIndexChanged.connect(self._quantity_menu_changed_action)
            self._quantity_menu.lineEdit().editingFinished.connect(self._quantity_menu_changed_action)
            self._quantity_menu.adjustSize()
            self.setWindowTitle("topsy: "+self._visualizer.data_loader.get_filename())

        self.call_later(0, populate_quantity_menu)

        self._toolbar.addAction(self._load_script_action)
        self._toolbar.addAction(self._save_script_action)
        self._toolbar.addAction(self._record_action)
        self._toolbar.addAction(self._save_movie_action)

        self._toolbar.addSeparator()
        self._toolbar.addAction(self._save_action)
        self._toolbar.addSeparator()
        self._toolbar.addWidget(self._colormap_menu)
        self._toolbar.addWidget(self._quantity_menu)
        self._toolbar.addSeparator()
        self._toolbar.addAction(self._link_action)
        self._recorder = None



        # now replace the wgpu layout with our own
        layout = self.layout()
        layout.removeWidget(self._subwidget)

        our_layout = PySide6.QtWidgets.QVBoxLayout()
        our_layout.addWidget(self._subwidget)
        our_layout.addWidget(self._toolbar)
        our_layout.setContentsMargins(0, 0, 0, 0)
        our_layout.setSpacing(0)

        self._toolbar.adjustSize()

        self._toolbar_update_timer = QtCore.QTimer(self)
        self._toolbar_update_timer.timeout.connect(self._update_toolbar)
        self._toolbar_update_timer.start(100)

        layout.addLayout(our_layout)

    def __del__(self):
        try:
            self._all_instances.remove(self)
        except ValueError:
            pass
        super().__del__()
    def _load_icons(self):
        self._record_icon = _get_icon("record.png")
        self._stop_icon = _get_icon("stop.png")
        self._save_icon = _get_icon("camera.png")
        self._linked_icon = _get_icon("linked.png")
        self._unlinked_icon = _get_icon("unlinked.png")
        self._save_movie_icon = _get_icon("movie.png")
        self._import_icon = _get_icon("load_script.png")
        self._export_icon = _get_icon("save_script.png")

    def _colormap_menu_changed_action(self):
        logger.info("Colormap changed to %s", self._colormap_menu.currentText())
        self._visualizer.colormap_name = self._colormap_menu.currentText()

    def _quantity_menu_changed_action(self):
        logger.info("Quantity changed to %s", self._quantity_menu.currentText())
        if self._quantity_menu.currentText() == self._default_quantity_name:
            self._visualizer.quantity_name = None
        else:
            try:
                self._visualizer.quantity_name = self._quantity_menu.currentText()
            except ValueError as e:
                message = QtWidgets.QMessageBox(self)
                message.setWindowTitle("Invalid quantity")
                message.setText(str(e))
                message.setIcon(QtWidgets.QMessageBox.Icon.Critical)
                message.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ok)
                self._quantity_menu.setCurrentText(self._visualizer.quantity_name or self._default_quantity_name)
                message.exec()


    def on_click_record(self):

        if self._recorder is None or not self._recorder.recording:
            logger.info("Starting recorder")
            self._recorder = VisualizationRecorderWithQtProgressbar(self._visualizer, self)
            self._recorder.record()
            self._record_action.setIconText("Stop")
            self._record_action.setIcon(self._stop_icon)
        else:
            logger.info("Stopping recorder")
            self._recorder.stop()
            self._record_action.setIconText("Record")
            self._record_action.setIcon(self._record_icon)

    def on_click_save_movie(self):
        # show the options dialog first:
        dialog = RecordingSettingsDialog(self)
        dialog.exec()
        if dialog.result() == QtWidgets.QDialog.DialogCode.Accepted:
            fd = QtWidgets.QFileDialog(self)
            fname, _ = fd.getSaveFileName(self, "Save video", "", "MP4 (*.mp4)")
            if fname:
                logger.info("Saving video to %s", fname)
                self._recorder.save_mp4(fname, show_colorbar=dialog.show_colorbar,
                                        show_scalebar=dialog.show_scalebar,
                                        fps=dialog.fps,
                                        resolution=dialog.resolution,
                                        smooth=dialog.smooth,
                                        set_vmin_vmax=dialog.set_vmin_vmax,
                                        set_quantity=dialog.set_quantity)
                QtGui.QDesktopServices.openUrl(QtCore.QUrl.fromLocalFile(fname))

    def on_click_save(self):
        fd = QtWidgets.QFileDialog(self)
        fname, _ = fd.getSaveFileName(self, "Save snapshot", "", "PNG (*.png);; PDF (*.pdf)")
        if fname:
            logger.info("Saving snapshot to %s", fname)
            self._visualizer.save(fname)
            QtGui.QDesktopServices.openUrl(QtCore.QUrl.fromLocalFile(fname))

    def on_click_save_script(self):
        fd = QtWidgets.QFileDialog(self)
        fname, _ = fd.getSaveFileName(self, "Save camera movements", "", "Python Pickle (*.pickle)")
        if fname:
            logger.info("Saving timestream to %s", fname)
            self._recorder.save_timestream(fname)

    def on_click_load_script(self):
        fd = QtWidgets.QFileDialog(self)
        fname, _ = fd.getOpenFileName(self, "Load camera movements", "", "Python Pickle (*.pickle)")
        if fname:
            logger.info("Loading timestream from %s", fname)
            self._recorder = VisualizationRecorderWithQtProgressbar(self._visualizer, self)
            self._recorder.load_timestream(fname)


    def on_click_link(self):
        if self._visualizer.is_synchronizing():
            logger.info("Stop synchronizing")
            self._visualizer.stop_synchronizing()
        else:
            logger.info("Start synchronizing")
            from .. import view_synchronizer
            synchronizer = view_synchronizer.ViewSynchronizer()
            for instance in self._all_instances:
                synchronizer.add_view(instance._visualizer)

    def _update_toolbar(self):
        if self._recorder is not None or len(self._all_instances)<2:
            self._link_action.setDisabled(True)
        else:
            self._link_action.setDisabled(False)
            if self._visualizer.is_synchronizing():
                self._link_action.setIcon(self._linked_icon)
                self._link_action.setIconText("Unlink")
            else:
                self._link_action.setIcon(self._unlinked_icon)
                self._link_action.setIconText("Link")
        if self._recorder is not None and not self._recorder.recording:
            self._save_movie_action.setDisabled(False)
            self._save_script_action.setDisabled(False)
        else:
            self._save_movie_action.setDisabled(True)
            self._save_script_action.setDisabled(True)




    def request_draw(self, function=None):
        # As a side effect, wgpu gui layer stores our function call, to enable it to be
        # repainted later. But we want to distinguish such repaints and handle them
        # differently, so we need to replace the function with our own
        def function_wrapper():
            function()
            self._subwidget.draw_frame = lambda: self._visualizer.draw(DrawReason.PRESENTATION_CHANGE)

        super().request_draw(function_wrapper)

    @classmethod
    def call_later(cls, delay, fn, *args):
        call_later(delay, fn, *args)



from __future__ import annotations

import numpy as np
import wgpu.gui.jupyter, wgpu.gui.auto

from ..drawreason import DrawReason

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..visualizer import Visualizer


from wgpu.gui.base import WgpuCanvasBase


class VisualizerCanvasBase:
    def __init__(self, *args, **kwargs):
        self._visualizer : Visualizer = kwargs.pop("visualizer")

        self._last_x = 0
        self._last_y = 0
        # The below are dummy values that will be updated by the initial resize event
        self.width_physical, self.height_physical = 640,480
        self.pixel_ratio = 1

        super().__init__(*args, **kwargs)

    def handle_event(self, event):
        if event['event_type']=='pointer_move':
            if len(event['buttons'])>0:
                if len(event['modifiers'])==0:
                    self.drag(event['x']-self._last_x, event['y']-self._last_y)
                else:
                    self.shift_drag(event['x']-self._last_x, event['y']-self._last_y)
            self._last_x = event['x']
            self._last_y = event['y']
        elif event['event_type']=='wheel':
            self.mouse_wheel(event['dx'], event['dy'])
        elif event['event_type']=='key_up':
            self.key_up(event['key'])
        elif event['event_type']=='resize':
            self.resize_complete(event['width'], event['height'], event['pixel_ratio'])
        elif event['event_type']=='double_click':
            self.double_click(event['x'], event['y'])
        elif event['event_type']=='pointer_up':
            self.release_drag()
        else:
            pass
        super().handle_event(event)

    def drag(self, dx, dy):
        self._visualizer.rotate(dx*0.01, dy*0.01)

    def shift_drag(self, dx, dy):
        biggest_dimension = max(self.width_physical, self.height_physical)

        displacement = 2.*self.pixel_ratio*np.array([dx, -dy, 0], dtype=np.float32) / biggest_dimension * self._visualizer.scale
        self._visualizer.position_offset += self._visualizer.rotation_matrix.T @ displacement

        self._visualizer.display_status("centre = [{:.2f}, {:.2f}, {:.2f}]".format(*self._visualizer._sph.position_offset))

        self._visualizer.crosshairs_visible = True


    def key_up(self, key):
        if key=='s':
            self._visualizer.save()
        elif key=='r':
            self._visualizer.vmin_vmax_is_set = False
            self._visualizer.invalidate()
        elif key=='h':
            self._visualizer.reset_view()

    def mouse_wheel(self, delta_x, delta_y):
        if isinstance(self, wgpu.gui.jupyter.JupyterWgpuCanvas):
            # scroll events are much smaller from the web browser, for
            # some reason, compared with native windowing
            delta_y *= 10
            delta_x *= 10

        self._visualizer.scale*=np.exp(delta_y/1000)

    def release_drag(self):
        if self._visualizer.crosshairs_visible:
            self._visualizer.crosshairs_visible = False
            self._visualizer.invalidate()


    def resize(self, *args):
        # putting this here as a reminder that the resize method must be passed to the base class
        super().resize(*args)
    def resize_complete(self, width, height, pixel_ratio=1):
        self.width_physical = int(width*pixel_ratio)
        self.height_physical = int(height*pixel_ratio)
        self.pixel_ratio = pixel_ratio

    def double_click(self, x, y):
        pass

    @classmethod
    def call_later(cls, delay, fn, *args):
        raise NotImplementedError()





# Now we are going to select a specific backend
#
# we don't use wgpu.gui.auto directly because it prefers the glfw backend over qt
# whereas we want to use qt
#
# Note also that is_jupyter as implemented fails to distinguish correctly if we are
# running inside a kernel that isn't attached to a notebook. There doesn't seem to
# be any way to distinguish this, so we live with it for now.
from wgpu.gui.auto import is_jupyter


if is_jupyter():
    from .jupyter import VisualizerCanvas
else:
    from .qt import VisualizerCanvas




"""topsy - An astrophysics simulation visualization package based on webgpu, using pynbody for reading data"""

from __future__ import annotations

__version__ = "0.3.3"

import argparse
import logging
import sys

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import pynbody


from . import config


def parse_args(args=None):
    """Create arguments and kwargs to pass to the visualizer, from sys.argv"""

    # create the argument parser, add arguments for filename, resolution, and colormap, and parse the arguments
    argparser = argparse.ArgumentParser(description="Visualize an astrophysics simulation. Multiple windows can be opened by separating groups of arguments with *.")

    argparser.add_argument("filename", help="Specify path to a simulation file to be visualized")
    argparser.add_argument("--resolution", "-r", help="Specify the resolution of the visualization",
                           default=config.DEFAULT_RESOLUTION, type=int)
    argparser.add_argument("--colormap", "-m", help="Specify the matplotlib colormap to be used",
                           default=config.DEFAULT_COLORMAP, type=str)
    argparser.add_argument("--particle", "-p", help="Specify the particle type to visualise",
                            default="dm", type=str)
    argparser.add_argument("--center", "-c", help="Specify the centering method: 'halo-<N>', 'all', 'zoom' or 'none'",
                           default="halo-1", type=str)
    argparser.add_argument("--quantity", "-q", help="Specify a quantity to render instead of density",
                           default=None, type=str)
    argparser.add_argument("--tile", "-t", help="Wrap and tile the simulation box using its periodicity",
                           default=False, action="store_true")

    if args is None:
        args = sys.argv[1:]
    arg_batches = []
    # split args into batches separated by '+'
    while len(args) > 0:
        try:
            split_index = args.index("+")
        except ValueError:
            split_index = len(args)
        arg_batches.append(args[:split_index])
        args = args[split_index+1:]

    return [argparser.parse_args(batch) for batch in arg_batches]

def setup_logging():
    global logger
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(
        logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    )
    logger.addHandler(ch)

def main():
    setup_logging()
    all_args = parse_args()

    from . import visualizer, loader

    for args in all_args:
        if "test://" in args.filename:
            loader_class = loader.TestDataLoader
            try:
                n_part = int(float(args.filename[7:])) # going through float allows scientific notation
            except ValueError:
                n_part = config.TEST_DATA_NUM_PARTICLES_DEFAULT
            logger.info(f"Using test data with {n_part} particles")
            loader_args = (n_part,)
        else:
            loader_class = loader.PynbodyDataLoader
            loader_args = (args.filename, args.center, args.particle)


        vis = visualizer.Visualizer(data_loader_class=loader_class,
                                    data_loader_args=loader_args,
                                    colormap_name=args.colormap,
                                    periodic_tiling=args.tile,
                                    render_resolution=args.resolution)

        vis.quantity_name = args.quantity
        vis.canvas.show()

    from wgpu.gui import qt
    qt.run()

def topsy(snapshot: pynbody.snapshot.SimSnap, quantity: str | None = None, **kwargs):
    from . import visualizer, loader
    vis = visualizer.Visualizer(data_loader_class=loader.PynbodyDataInMemory,
                                data_loader_args=(snapshot,),
                                **kwargs)
    vis.quantity_name = quantity
    return vis

def _test(nparticle=config.TEST_DATA_NUM_PARTICLES_DEFAULT, **kwargs):
    from . import visualizer, loader
    vis = visualizer.Visualizer(data_loader_class=loader.TestDataLoader,
                                data_loader_args=(nparticle,),
                                **kwargs)
    return vis


DEFAULT_RESOLUTION = 1024
DEFAULT_COLORMAP = 'twilight_shifted'

DEFAULT_SCALE = 200.0 # viewport width in kpc

TARGET_FPS = 30 # will use downsampling to achieve this
FULL_RESOLUTION_RENDER_AFTER = 0.3 # inactivity seconds to wait before rendering without downs
STATUS_LINE_UPDATE_INTERVAL = 0.2 # seconds
STATUS_LINE_UPDATE_INTERVAL_RAPID = 0.05 # when time-critical information is being displayed

COLORBAR_ASPECT_RATIO = 0.15
COLORMAP_NUM_SAMPLES = 1000

TEST_DATA_NUM_PARTICLES_DEFAULT = int(1e6)


from io import BytesIO

import matplotlib
from matplotlib.figure import Figure
import matplotlib.pyplot as plt

from .overlay import Overlay


class TextOverlay(Overlay):
    def __init__(self, visualizer, text, clipspace_origin, pixelspace_height, *, dpi=200, **kwargs):
        self.text = text
        self.dpi = dpi
        self.clipspace_origin = clipspace_origin
        self.pixelspace_height = pixelspace_height
        self.kwargs = kwargs

        super().__init__(visualizer)

    def get_clipspace_coordinates(self, width, height):
        im = self.get_contents()
        x,y = self.clipspace_origin
        height = self.pixelspace_height/height
        width = self.pixelspace_height*im.shape[1]/im.shape[0]/width
        return x, y, width, height

    def render_contents(self):
        return self.text_to_rgba(self.text, dpi=self.dpi, **self.kwargs)

    @staticmethod
    def text_to_rgba(s, *, dpi, **kwargs):
        """Render text to RGBA image.

        Based on
        https://matplotlib.org/stable/gallery/text_labels_and_annotations/mathtext_asarray.html"""
        fig = Figure(facecolor="none")
        fig.text(0, 0, s, **kwargs)
        with BytesIO() as buf:
            fig.savefig(buf, dpi=dpi, format="png", bbox_inches="tight",
                        pad_inches=0)
            buf.seek(0)
            rgba = plt.imread(buf)
        return rgba


from __future__ import annotations

import time
import tqdm
import wgpu
import numpy as np
import logging
import pickle

from . interpolator import (Interpolator, StepInterpolator, LinearInterpolator, RotationInterpolator,
                            SmoothedRotationInterpolator, SmoothedLinearInterpolator, SmoothedStepInterpolator)
from ..drawreason import DrawReason
from ..view_synchronizer import ViewSynchronizer

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..visualizer import Visualizer

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)



class VisualizationRecorder:
    _record_properties = ['colormap_name', 'quantity_name', 'log_scale', 'vmin', 'vmax', # NB ordering is important to prevent triggering auto-scaling
                          'rotation_matrix', 'scale', 'position_offset']
    _record_interpolation_class_smoothed = [StepInterpolator, StepInterpolator, StepInterpolator, SmoothedStepInterpolator, SmoothedStepInterpolator,
                                   SmoothedRotationInterpolator, SmoothedLinearInterpolator, SmoothedLinearInterpolator]
    _record_interpolation_class_unsmoothed = [StepInterpolator, StepInterpolator, StepInterpolator, StepInterpolator,
                                   StepInterpolator, RotationInterpolator, LinearInterpolator, LinearInterpolator]


    def __init__(self, visualizer: Visualizer):
        vs = ViewSynchronizer(synchronize=self._record_properties)
        vs.add_view(visualizer)
        vs.add_view(self)
        self._recording = False
        self._playback = False
        self._recording_ends_at = None
        self._visualizer = visualizer
        self._reset_timestream()

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if key in self._record_properties:
            self._view_synchronizer.update_completed(self)  # this marks the update as done
            if self._recording:
                self._timestream[key].append((self._time_elapsed(), value))

    def _time_elapsed(self):
        return time.time() - self._t0

    def _reset_timestream(self):
        self._timestream = {r: [(0.0, getattr(self._visualizer, r))] for r in self._record_properties}

    def record(self):
        self._t0 = time.time()
        self._reset_timestream()
        self._recording = True
        self._playback = False

    def stop(self):
        if self._recording:
            self._recording_ends_at = self._time_elapsed()
        self._recording = False
        self._playback = False


    def _get_value_at_time(self, property, time):
        return self._interpolators[property](time)

    def _progress_iterator(self, ntot):
        """Return an iterator that displays progress in an appropriate way

        Overriden for the qt gui"""
        return tqdm.tqdm(range(ntot), unit="frame")

    def _replay(self, fps=30.0, resolution=(1920, 1080), show_colorbar=True,
                show_scalebar=True, smooth=True, set_vmin_vmax=True,
                set_quantity=True):
        if self._recording:
            self.stop()
        if self._recording_ends_at is None:
            raise RuntimeError("Can't playback before recording")

        self._recording = False
        self._playback = True

        exclude = []

        if not set_vmin_vmax:
            exclude.extend(['vmin', 'vmax'])
        if not set_quantity:
            exclude.append('quantity_name')


        try:
            self._visualizer.show_colorbar = show_colorbar
            self._visualizer.show_scalebar = show_scalebar
            if smooth:
                self._interpolators = {r: c(self._timestream[r])
                                       for c, r in zip(self._record_interpolation_class_smoothed,
                                                       self._record_properties)
                                       if r not in exclude}
            else:
                self._interpolators = {r: c(self._timestream[r])
                                       for c, r in zip(self._record_interpolation_class_unsmoothed,
                                                       self._record_properties)
                                       if r not in exclude}

            device = self._visualizer.device

            render_texture: wgpu.GPUTexture = device.create_texture(
                size=(resolution[0], resolution[1], 1),
                usage=wgpu.TextureUsage.RENDER_ATTACHMENT |
                      wgpu.TextureUsage.COPY_SRC,
                format=self._visualizer.canvas_format,
                label="output_texture",
            )

            num_frames = int(self._recording_ends_at * fps)
            for i in self._progress_iterator(num_frames):
                t = i / fps
                for p in self._record_properties:
                    if p not in exclude:
                        val = self._get_value_at_time(p, t)
                        if val is not Interpolator.no_value:
                            setattr(self._visualizer, p, val)

                self._visualizer.display_status("github.com/pynbody/topsy/")
                self._visualizer.draw(DrawReason.EXPORT, render_texture.create_view())
                im = device.queue.read_texture({'texture': render_texture, 'origin': (0, 0, 0)},
                                               {'bytes_per_row': 4 * resolution[0]},
                                               (resolution[0], resolution[1], 1))
                im_npy = np.frombuffer(im, dtype=np.uint8).reshape((resolution[1], resolution[0], 4))
                im_npy = im_npy[:, :,:3]
                yield im_npy

            self.playback = False
        finally:
            self._visualizer.show_colorbar = True
            self._visualizer.show_scalebar = True

    def save_mp4(self, filename, fps, resolution, *args, **kwargs):
        import cv2
        writer = cv2.VideoWriter(filename, cv2.VideoWriter.fourcc(*'mp4v'), fps,
                                 resolution)

        for image in self._replay(fps, resolution, *args, **kwargs):
            writer.write(image)

        writer.release()

    def save_timestream(self, fname):
        pickle.dump((self._timestream, self._recording_ends_at), open(fname, 'wb'))

    def load_timestream(self, fname):
        self._timestream, self._recording_ends_at = pickle.load(open(fname, 'rb'))


    @property
    def recording(self):
        return self._recording




from wgpu.gui.jupyter import WgpuCanvas, call_later
from . import VisualizerCanvasBase

class VisualizerCanvas(VisualizerCanvasBase, WgpuCanvas):
    def request_draw(self, function=None):

        # As a side effect, wgpu gui layer stores our function call, to enable it to be
        # repainted later. But we want to distinguish such repaints and handle them
        # differently, so we need to replace the function with our own

        #def function_wrapper():
        #    function()
        #    self._subwidget.draw_frame = lambda: self._visualizer.draw(DrawReason.PRESENTATION_CHANGE)
        # TODO: above needs to be hacked for jupyter

        super().request_draw(function)

    @classmethod
    def call_later(cls, delay, fn, *args):
        call_later(delay, fn, *args)







from __future__ import annotations

import weakref

from .drawreason import DrawReason

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .visualizer import Visualizer


class ViewSynchronizer:
    """A class to manage reflecting rotation/scale/offset changes from one view to another.

    Best used by creating two visualizers, visualizer1 and visualizer2, and then calling
    visualizer1.synchronize_with(visualizer2)

    The two visualizers will then be kept in sync
    """
    def __init__(self, synchronize=['rotation_matrix', 'scale', 'position_offset']):
        self._views : list[weakref.ReferenceType[Visualizer]] = []
        self._requires_update : list[weakref.ReferenceType[Visualizer]] = []
        self._synchronize = synchronize

    def perpetuate_update(self, source):
        """Called when a view has been updated and the update needs to be perpetuated to other views.

        Note that there is built-in protection against infinite loops. If view A calls this method
        which causes view B to update, if view B calls this method nothing will be issued back to view A."""
        sources_needing_update = [view_weakref() for view_weakref in self._requires_update]
        if source in sources_needing_update:
            # OK the update has happened! Great, but don't broadcast it again
            del self._requires_update[sources_needing_update.index(source)]
            return

        for view_weakref in self._views:
            view = view_weakref()
            if (view is not source and view is not None) and (view_weakref not in self._requires_update):
                self._requires_update.append(view_weakref)
                for var in self._synchronize:
                    setattr(view, var, getattr(source, var))

    def update_completed(self, view: Visualizer):
        """Called when a view knows it will not be attempting to perpetuate an update it received

        See note about infinite loops in perpetuate_update. This method is used when a view will
        not be acting on the update it received, so it must be removed from the exclusion list
        that perpetuate_update maintains."""
        sources_needing_update = [view_weakref() for view_weakref in self._requires_update]
        if view in sources_needing_update:
            del self._requires_update[sources_needing_update.index(view)]

    def add_view(self, view: Visualizer):
        self._views.append(weakref.ref(view))
        view._view_synchronizer = self

    def remove_view(self, view: Visualizer):
        self._views.remove(weakref.ref(view))
        del view._view_synchronizer

class SynchronizationMixin:
    """Mixin class for Visualizer to allow it to synchronize with other views"""
    def draw(self, reason, render_texture_view=None):
        super().draw(reason, render_texture_view)
        if hasattr(self, "_view_synchronizer") and reason != DrawReason.REFINE:
            self._view_synchronizer.perpetuate_update(self)

    def synchronize_with(self, other: Visualizer):
        """Start synchronizing this visualizer with another"""
        if hasattr(self, "_view_synchronizer") and hasattr(other, "_view_synchronizer"):
            raise RuntimeError("Both these visualizers are already synchronizing with others")

        if hasattr(self, "_view_synchronizer"):
            self._view_synchronizer.add_view(other)
        elif hasattr(other, "_view_synchronizer"):
            other._view_synchronizer.add_view(self)
        else:
            vs = ViewSynchronizer()
            vs.add_view(self)
            vs.add_view(other)

    def stop_synchronizing(self):
        """Stop synchronizing this visualizer with any other"""
        if hasattr(self, "_view_synchronizer"):
            self._view_synchronizer.remove_view(self)

    def is_synchronizing(self):
        return hasattr(self, "_view_synchronizer")
